% Prolog implementation of definitions in the paper:
% What can I do to help you?
% A formal framework for agent reasoning about behavior change support for people
% 
% Tested with GNU-Prolog
%
% Usage: go.

go :-
    example2_4, nl, 
    example2_7, nl,
    example2_9,  nl,
    example3_2, nl,
    example3_4, nl,
    example3_8,
    write('Done.\n\n').

% In Prolog more natural to represent the tree directly as a nested term.

% An activity tree is of the form: (see eg_car below)
% T ::= leaf(N) | or(N,[Children]) | xor(N,[Children]) | opt(N, [MandatoryChildren], [NonMandatoryChildren])

% edge(T,N,M,L) iff T has an edge labelled L from N to M
edge(T,N,M,L) :- T=or(N,Cs), member(M1,Cs), getRoot(M1,M), L=or.
edge(T,N,M,L) :- T=xor(N,Cs), member(M1,Cs), getRoot(M1,M), L=xor.
edge(T,N,M,L) :- T=opt(N,Cs,_), member(M1,Cs), getRoot(M1,M), L=mandatory.
edge(T,N,M,L) :- T=opt(N,_,Cs), member(M1,Cs), getRoot(M1,M), L=optional.
edge(T,N,M,L) :- getKids(T,Cs), member(T1,Cs), edge(T1,N,M,L).

% nodes(T,Ns) iff Ns is the list of nodes in the tree T
nodes(T,[Root|Nodes]) :- getRoot(T,Root), setof(N, M^L^edge(T,M,N,L), Nodes).

% getRoot(T,N) iff n is the root of the tree T
getRoot(or(N,_),N).
getRoot(xor(N,_),N).
getRoot(opt(N,_,_),N).
getRoot(leaf(N),N).

% getKids(T,C) is C is the list of nodes that are the children of the top node in T
getKids(or(_,C),C1) :- C1=C.
getKids(xor(_,C),C1) :-  C1=C.
getKids(opt(_,C1,C2),C12) :- append(C1,C2,C12).

% Example in Figure 1 our notes ... v10: updated terminology and structure to latest version
eg_car( xor(going_to_work, [By_bike, By_car]) ) :- 
        By_bike = opt(by_bike, 
            [leaf(get_bike_keys), 
             leaf(bike_route_to_work)], [leaf(put_on_raincoat)]), 
        By_car = opt(by_car, 
            [leaf(get_car_keys), leaf(drive_route_to_work)], 
            [leaf(navigation)]).

% List of nodes that are ambiguous 
amb(N) :- member(N, [get_bike_keys, put_on_raincoat, get_car_keys,navigate]).

% Definition 2.3: practical model
% model M is practical for T if there are no invalid edges (last condition), 
% and practicalAux holds (which checks the first three conditions)
practical(T,M) :-  \+ invalidEdge(M,T), practicalAux(M,T).

% invalidEdge(M,T) iff M is not a practical model for T because it violates
% the last condition: there is an edge from N to M where M is in the model, but
% N is not in the model, and M is not ambiguous. 
invalidEdge(Model,T) :- edge(T,N,M,_), 
    \+ member(N,Model), member(M, Model), \+ amb(M).

% practicalAux(Model,Tree) iff the first three conditions of definition 2.3
% hold with respect to the model and tree
practicalAux(_,leaf(_)). 
practicalAux(M,or(N,Cs)) :- !, (member(N,M) -> (member(C,Cs), getRoot(C,R), member(R,M)) ; true), maplist(practicalAux(M),Cs).
practicalAux(M,xor(N,Cs)) :- !, (member(N,M) -> existsOne(C, (member(C,Cs), getRoot(C,R), member(R,M))) ; true), maplist(practicalAux(M),Cs).
practicalAux(M,opt(N,Cs,Cs2)) :- !, (member(N,M) -> forall(member(C,Cs), (getRoot(C,R), member(R,M))) ; true), maplist(practicalAux(M),Cs), maplist(practicalAux(M),Cs2).

% example model in Figure 1
eg_car_ab([going_to_work, by_car, get_car_keys, drive_route_to_work]).

% Example 2.4
example2_4 :- eg_car(T), eg_car_ab(Model), 
    write('Example 2.4: model '), write(Model), 
    (practical(T,Model) ->  write(' is practical.\n') ; 
                        write(' is NOT practical.\n')).

% A behaviour model (Definition 2.5) is represented in Prolog as 
% T, AB, DC, S
% where T is a tree (as above), AB is a model (list of nodes), 
% and DC = dcn(Change, Keep) where Change and Keep are lists of nodes.
% This is simpler than mapping each node to "keep" or "change"

% Figure 1: dc1 (start by_bike) and dc2 (stop by_car)
eg_car_dc1( dcn([by_bike],[]) ).
eg_car_dc2( dcn([by_car],[]) ).

% Definition 2.6
% delta(AB,DC,AB2) iff AB2 is the result of applying DC to AB
delta(AB, dcn(Change,_), ABp) :- changeadd(Change,AB,ABp).
changeadd([],AB,AB).
changeadd([X|Xs],AB,AB2) :- 
	member(X,AB) -> (delete(AB,X,AB1), changeadd(Xs,AB1,AB2)) ; (AB2=[X|AB1], changeadd(Xs,AB,AB1)).

% deltastar(T, AB, DC, DS) iff DS is the list of models according to Delta^*
% Note that S is not needed, so is not passed in
% The definition finds minimal repair options, and for each applies delta
% with the combination of the changes in DC and the MRO
% Note that setof fails if the condition is false, hence the check to make 
% deltastar return an empty list in this case
deltastar(T, AB, DC, DS) :- DC=dcn(Change,Keep),
	(setof(DS, MRO^ChangeMRO^(minimalRepairOption(T, AB, DC, MRO), 
		append(Change,MRO,ChangeMRO),
		delta(AB,dcn(ChangeMRO,Keep), DS)), DStar)
     -> DS=DStar; DS=[]).

% a minimal repair option for T, AB, DC is a repairOption that is minimal, 
% i.e. all (proper) sublists of the minimal repair option are not repair options
minimalRepairOption(T, AB, DC, MRO) :- repairOption(T, AB, DC, MRO), 
  forall( (sublist(MROp,MRO), (MROp\=MRO)), \+ repairOption(T,AB,DC,MROp)).

% a repair option is constructed out of things that are in the tree
% but not in DC, and applying it AB together with the changes in DC (using delta) 
% gives a practical model
% note: since practical/2 can succeed multiple times, need once(practical(...))
repairOption(T, AB, dcn(Change,Keep), RO) :- 
	nodes(T,Ns), subtract(Ns,Change,Ns1), subtract(Ns1,Keep,Ns2),
        sublist(RO,Ns2), append(Change,RO,ChangeRO),
	delta(AB, dcn(ChangeRO,Keep), ABp), once(practical(T,ABp)).

% Example 2.7
example2_7 :- eg_car(T), eg_car_ab(AB), 
    DC = dcn([by_bike],[going_to_work]),
    delta(AB,DC,AB2), 
    write('Example 2.7: delta(BM) = '), 
    write(AB2), write(' which '), 
    (practical(T,AB2) ->  write(' is practical.\n') ; 
                        write(' is NOT practical.\n')),
    deltastar(T, AB, DC, DStar), 
    write('delta* = '), write(DStar), nl.

% Definition 2.8
achievable(T, AB, DC) :- deltastar(T, AB, DC, DS), DS \= [].

% Example 2.9 
example2_9 :- 
    write('Example 2.9: '), 
    eg_car(T), eg_car_ab(AB), 
    DC = dcn([by_bike],[going_to_work,by_car]),
    deltastar(T, AB, DC, DStar), 
    write(' delta* = '), write(DStar), nl.

% Definition 2.10
% a DC holds with respect to an original AB and changed AB if: 
% (i) everything that should change has changed, and 
% (ii) everything that should be kept has not changed.
dc_valid(DC,AB,AB2) :- DC=dcn(Change,Keep), 
	forall(member(C,Keep), kept(C,AB,AB2)),
	forall(member(C,Change), changed(C,AB,AB2)).
kept(C,AB,AB2) :- (member(C,AB), member(C,AB2)) ; ((\+member(C,AB)), (\+member(C,AB2))).
changed(C,AB,AB2) :- (member(C,AB), (\+member(C,AB2))) ; ((\+member(C,AB)), member(C,AB2)).

% Definition 3.1
% cause(T, AB, DC, S, Cau) iff Cau is a change set for the BM = (T, AB, DC, S)
% Note that rather than return a set of cause sets, cause returns a single cause
% set, with further answers on backtracking.
cause(T,AB,DC,S, Cau) :- DC=dcn(Change,Keep),
    % Build a candiate Cause set Cau
	nodes(T,Nodes), subtract(Nodes,Change,N1), subtract(N1,Keep,CandidateCs),
	sublist(Cau,CandidateCs),
    % compute DC_ns = dcn(ChangeS, KeepS), and CauS = Cau \cup DC_ns
	filter_s(Change,S,ChangeS), filter_s(Keep,S,KeepS), append(Cau,ChangeS,CauS),
    % compute DC_s = dcn(Change_NS, Keep_NS)
	filter_ns(Change,S,Change_NS), filter_ns(Keep,S,Keep_NS), 
    DC_s = dcn(Change_NS,Keep_NS),
	deltastar(T, AB, dcn(CauS,KeepS), DStar), 
	DStar \= [], 
	forall(member(DS, DStar), dc_valid(DC_s,AB,DS)), 
	\+((propersublist(CauPrime,Cau), append(CauPrime,ChangeS,CauPrimeS),
		deltastar(T, AB, dcn(CauPrimeS,KeepS), DStarP), 
		DStarP \= [],
		forall(member(DSP,DStarP), dc_valid(DC_s,AB,DSP)))).

% Filter out what is in S, i.e. keep only what is NOT in S
filter_s([],_,[]).
filter_s([N|Ns],S,[N|R]) :- \+ member(N,S), filter_s(Ns,S,R).
filter_s([N|Ns],S,R) :- member(N,S), filter_s(Ns,S,R).

% Filter out what is NOT in S, i.e. keep only what IS in S
filter_ns([],_,[]).
filter_ns([N|Ns],S,[N|R]) :- member(N,S), filter_ns(Ns,S,R).
filter_ns([N|Ns],S,R) :- \+ member(N,S), filter_ns(Ns,S,R).

propersublist(X,Y) :- sublist(X,Y), X\=Y.

% Example 3.2
example3_2 :- write('Example 3.2: '), nl,
    eg_car(T), eg_car_ab(AB), eg_car_dc1(DC), S = [by_bike],
    cause(T, AB, DC, S, Cau), 
    write(' ... Cau = '), write(Cau), nl,
    fail.
example3_2 :- write('...Done part 1 ...\n'), fail.
example3_2 :- write('Example 3.2: (part 2) '), nl,
    eg_car(T), eg_car_ab(AB), S = [by_bike],
    DC = dcn([by_bike],[going_to_work]),
    cause(T, AB, DC, S, Cau), 
    write(' ... Cau = '), write(Cau), nl,
    fail.
example3_2 :- write('...Done part 2.\n').

% Definition 3.3
% pob(T, AB, DC, S, POB) iff POB is a possible obstacle for the BM=(T,AB,DC,S)
% Note that pob returns a single POB and is backtrackable 
pob(T,AB,DC,S, POB) :-  DC=dcn(Change,Keep),
    % Find a candidate POB
	nodes(T,Nodes), subtract(Nodes,Change,N1), subtract(N1,Keep,CandidateCs),
	member(POB,CandidateCs),
    % First condition
	deltastar(T, AB, DC, DStar), 
	forall( member(DS, DStar) , dc_valid(dcn([POB],[]), AB, DS) ), 
    % Second condition, DC_s = dcn(Change_NS, Keep_NS)
	filter_ns(Change,S,Change_NS), filter_ns(Keep,S,Keep_NS),
	deltastar(T, AB, dcn(Change_NS,Keep_NS), DStar2), 
    % use once to avoid repeated solutions
	once((member(DS2,DStar2), dc_valid(dcn([POB],[]), AB, DS2))).

% Example 3.4
example3_4 :- write('Example 3.4: \n'),
    eg_car(T), eg_car_ab(AB), eg_car_dc1(DC), S=[by_bike], 
    pob(T, AB, DC, S, POB), 
    write('... POB = '), write(POB), nl, 
    fail.
example3_4 :- write('... done.\n').
 
% Definition 3.5
% rcm(T, AB, DC, RCM) iff RCM is a member of the RCM set
rcm(_T, _AB, DC,_, rcm(A,direct)) :- DC=dcn(Change,_Keep), member(A,Change).
rcm(T, AB, DC, S, rcm(AM,indirectcau)) :- cause(T, AB, DC, S, AM).
rcm(T, AB, DC, S, rcm(AM,indirectpob)) :- pob(T, AB, DC,S, AM).

% Example 3.8
example3_8 :- write('Example 3.8: '), nl,
    eg_car(T), eg_car_ab(AB), 
    eg_car_dc1(StartByBike), eg_car_dc2(StopByCar),
    BothChanges = dcn([by_bike, by_car],[]),
    setof(RCM11, rcm(T,AB, StartByBike, [by_bike], RCM11), RCM1),
    setof(RCM21, rcm(T,AB, StopByCar, [by_car], RCM21), RCM2),
    setof(RCM31, rcm(T,AB, BothChanges, [by_bike,by_car], RCM31), RCM3),
    write('rcm for start by_bike = \n'), write_rcm(RCM1), nl,
    write('rcm for stop by_car = \n'), write_rcm(RCM2), nl,
    write('rcm for both = \n'), write_rcm(RCM3), nl.

write_rcm([]).
write_rcm([A|As]) :- write(' '), write(A), nl, write_rcm(As).


%%%% Library predicate
existsOne(X,Y) :- findall(X,Y,[_]).

%% Additional examples - generating tables for the supplementary material
%% Note: need to manually replace "_" with "\_" for LaTeX

% Columns = list of DC
table(AB,Columns) :-
    eg_car(T), 
    nodes(T,Ns),
    write('\\begin{tabular}{l'), 
    forall(member(dcn(_Change,_Keep), Columns), write('l')), 
    write('}\n\\hline\n'),
    write('DC Change: '), 
    forall(member(dcn(Change,_Keep), Columns), (write('& '), write(Change))), 
    write('\\\\\n'),
    write('Keep: '), 
    forall(member(dcn(_Change,Keep), Columns), (write('& '), write(Keep))), 
    write('\\\\\\hline\n'),
    forall(member(N,Ns), do_row(T,AB,N,Columns)),
    write('\\hline\n\\end{tabular}\n').

do_row(T,AB,N,Columns) :- 
    write(N),
    forall(member(DC,Columns), 
        (write(' & '), do_entry(T,AB,N,DC))), 
    write('\\\\\n').

do_entry(T,AB,N,DC) :- DC=dcn(Change,_Keep),
    (member(N,Change) -> write(' direct ') ;
     (cause(T,AB,DC,Change, Cau),member(N,Cau)) -> 
       (pob(T,AB,DC,Change,N) -> write(' cause+pob ') ; write(' cause ')) ;
       (pob(T,AB,DC,Change,N) -> write(' pob ') ; write(' - '))).
    
table1 :- 
    eg_car_ab(AB),
    table(AB, [ 
        dcn([by_bike],[]), dcn([by_bike],[going_to_work]),
        dcn([by_car],[]), dcn([by_car],[going_to_work]) ]).

table2 :- 
    AB=[going_to_work, by_bike, get_bike_keys, bike_route_to_work],
    table(AB, [ 
        dcn([by_bike],[]), dcn([by_bike],[going_to_work]),
        dcn([by_car],[]), dcn([by_car],[going_to_work]) ]).
    
    
